/*
 *  Parallelized EPRSD compressor of memory traces
 *  Created on Aug 25th 2009
 *  Author: Sandeep Budanur Ramanna
 *  E-mail: sbudanu@ncsu.edu
 *  North Carolina State University
 * */
#include <mpi.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include "types.h"
#include "myprsdtlib.h"

#define PRSD_PIPE "/home/sbudanu/pin-2.7-31933-gcc.3.4.6-ia32_intel64-linux/source/tools/MyMemTraceW"
#define BUF_LEN 600 //should be greater than equal to size of any data we are sending using MPI_Send()
#define MSG_TAG 99
#define MAIN_PROCESS 0
#define MAX_FILENAME_LEN 256
#define MAX_THREADS 32
#define MAX_TRACES 1000000

FILE* fpipe = NULL;
FILE* outfile = NULL;
_UINT64 sincelast = 0;
_UINT64 ip, ebp, sp, sign, thread_id, addr;
_UINT64 sign_size = 0;
_UINT64 sndSignArr[MAX_SIGN_LEN];
char type = 'x', valid = 'N';
char rcvBuffer[BUF_LEN];
char tracefilename[MAX_FILENAME_LEN];
struct timeval t1, t2;
time_t elapsed_time;

typedef enum
{
	STATUS_NONE = 0,
	STATUS_INTRATASK_COMP_IN_PROGRESS,
	STATUS_INTERTASK_COMP_IN_PROGRESS,
	STATUS_INTERNODE_COMP_IN_PROGRESS,  
	STATUS_MAX
}COMPSTATUS;

/* change the status to intratask, intertask and internode using the enum STATUS*/
COMPSTATUS compstatus =  STATUS_NONE;

/* msgrcvd[0] == 1 means intertask compression request msg has been received 
 * msgrcvd[1] == 1 means internode compression request msg has ben received
 * */
enum PRSD_COMP_MSGS_T
{
	//main process to child process messages, prefix MC_
	MC_TRACE, //msg having trace
	MC_TRACES_END, //msg indicating traces are over
	MC_RCV_FROM,//instruction to receive PRSDs from a specific process
	MC_SND_TO, //instructiion to send PRSDs to a specific process
	MC_SAVE_TO_FILE, //instruction to save PRSDs to a specified file
	MC_ERROR, //indicates ERROR
	MC_EXIT, //instruction for the child process to end
	MC_ACTIVE, //indicates that the main process is still active 

	//child process to main process messages, prefix CM_
	CM_QUERY_STATUS, //msg to 
	CM_READY_FOR_MERGE,
	CM_MERGE_DONE, //msg indicating the merge is completed with a specific process's PRSDs
	CM_PRSDS_SENT, //msg indiacting that all PRSDs were received
	CM_PRSDS_RCVD, //msg indicating that all PRSDs were sent

	//child process to child process messages, prefix CC_
	CC_PRSD, //msg having PRSD
	CC_PRSDS_END, //indicates PRSDs are over
	CC_PRSDS_REQ,	//request to send PRSDs, sent by receiver to sender
	CC_SIGTREE_NODE,
	CC_SIGTREE_REQ,
	CC_SIGTREE_END,
	/* msg types for internode compression */
	CC_SIGTREE_NODE_IN,
	CC_SIGTREE_REQ_IN,
	CC_SIGTREE_END_IN,
	CC_PRSD_IN,
	CC_PRSDS_END_IN,
	CC_PRSDS_REQ_IN,
	MSGS_MAX
};


/* example of deriving a class from abstract class INFO */
class MYINFO:public _INFO
{
	public:
	MYINFO():_INFO()
	{
	}

	MYINFO(MYINFO& src):_INFO()
	{
		//cout<<"\nMYINFO copy ctor called";
		value = src.value;
	}

	~MYINFO()
	{
		//cout <<"\nMYINFO dtor()";	
	}

	virtual _UINT64 getValue()	
	{
		//cout<<"\nMYINFO getValue called";
		return value;
	}		
	virtual void setValue(void* val)	
	{
		//cout<<"\nMYINFO setValue called";
		value = (_UINT64)val;
	}

};


/* class defining match function */
template <class T>
class MYMATCH
{
	public:
		/* compare function for intra-task compression */
		inline static bool compare(T* src, T* dest)
		{
			//nothing to compare - return true
			//returning false makes all EPRSDs to be appended - uncompressed case
			return true;
		}

		/* compare function for inter-thread compression */
		inline static bool interthread_compare(T* src, T* dest)
		{
			return true;
		}

		/* compare function for inter-node compression */
		inline static bool internode_compare(T* src, T* dest)
		{
			return true;
		}
};

/* merge function hook */
template <class T>
class MYMERGE
{
	public:
		/* template library merges the EPRSDs, this is just a notification to user - is it really required?*/
		/* EPRSDs are private to compressor object, cannot let user merge them. Library merges the EPRSDs after 
		   matching the signature and userdefined part in MYMATCH::compare()
		   */
		inline static void mergecb(T* src, T* dest)
		{	
			//cout<<endl<<"Merge_cb Src:"<<src->getValue()<<", "<<"Dest:"<<dest->getValue();	
			return;		
		}

		inline static void interthread_mergecb(T* src, T* dest)
		{	
			return;	
		}	
};

/* receiver sends PRSD request message to 'sender' telling that it is now ready to receive EPRSDs */
void sendPRSDRequestMsg_IN(int sender)
{
	char sndBuffer[BUF_LEN];
	int position = 0;
	int msg_type = CC_PRSDS_REQ_IN;
	MPI_Pack(&msg_type, 1, MPI_UNSIGNED, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
	MPI_Send(sndBuffer, BUF_LEN, MPI_PACKED, sender, MSG_TAG, MPI_COMM_WORLD);    
	return;
}

#ifdef SIGTREE_ENABLED
/* Internode trace compression - receiver sends the SIGTREE request message to 'sender' telling that it is now ready to receive signature tree */
void sendSigtreeRequestMsg_IN(int sender)
{
	char sndBuffer[BUF_LEN];
	int position = 0;
	int msg_type = CC_SIGTREE_REQ_IN;
	MPI_Pack(&msg_type, 1, MPI_UNSIGNED, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
	MPI_Send(sndBuffer, BUF_LEN, MPI_PACKED, sender, MSG_TAG, MPI_COMM_WORLD); 
	return;
}

/* Internode trace compression - sender sends this message to notify the receiver that signature tree transmission is complete */
void sendSigtreeEndMsg_IN(int dest)
{
	char sndBuffer[BUF_LEN];
	int position = 0;
	int msg_type = CC_SIGTREE_END_IN;
	MPI_Pack(&msg_type, 1, MPI_UNSIGNED, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
	MPI_Send(sndBuffer, BUF_LEN, MPI_PACKED, dest, MSG_TAG, MPI_COMM_WORLD);
	return;
}

/* Internode trace compression - send the whole signature tree - number the tree before sending*/
void sendSigtree_IN(int dest, EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE> *ecr, SIGTREE_ITERATOR *titer, _UINT64 mytid, _UINT64 myrank)
{
	char sndBuffer[BUF_LEN];
	int position = 0;
	int msg_type = CC_SIGTREE_NODE_IN;
	SIGTREEITEM item;
	printf("\nIN:Sending sig tree to %d", dest);
	ecr->numberSigTree(mytid, myrank);
	/* iterate the tree and send node by node */
	while(titer->next())
	{
		titer->currentItem(&item);
		//fprintf(stdout, "\n%4d, %4d, %4d, %20p ", item.data.key[0], item.data.key[1], item.data.key[2], item.data.info);
		//fprintf(stdout, "(%4d, %4d, %4d, %20p)", item.parentData.key[0], item.parentData.key[1], item.parentData.key[2], item.parentData.info);
		position = 0;
		MPI_Pack(&msg_type, 1, MPI_UNSIGNED, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&(item.data.key[0]), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&(item.data.key[1]), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&(item.data.key[2]), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&(item.data.info), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&(item.parentData.key[0]), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&(item.parentData.key[1]), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&(item.parentData.key[2]), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&(item.parentData.info), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Send(sndBuffer, BUF_LEN, MPI_PACKED, dest, MSG_TAG, MPI_COMM_WORLD);
		//getchar();
	}
	return;
}
#endif

/* sending all the EPRSDs in the compressor object to the receiver node for merging.
 * IMPORTANT: When signature tree is used, actual signatures are not sent instead, unique keys to
 * leaf nodes of the signature tree are sent in place of signature. These unique keys are resolved
 * on the receiver to point to the corresponding leaf node in the signature tree. */
void sendAllPRSDs_IN(int dest, EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE> *ecr, EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE>::ITERATOR *iter)
{
	char sndBuffer[BUF_LEN];
	int position = 0, i=0;
	int msg_type = CC_PRSD_IN;    	
	_INT64 value = 1122;
	_INT16 code = 0;
	_INT64 attr = ATTR_NONE;				
	EPRSD_DATA<MYINFO> *cur_prsd = NULL;
	int count = 0;
	int myrank = 0;	

	MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

	printf("sendAllPRSDs, from %d to dest: %d\n", myrank, dest); 
#ifdef SIGTREE_ENABLED
	/* copy unique keys of leaf nodes to the dependent EPRSDs */
	ecr->assignUniqueKeys();
#endif
	iter->begin();
	while(1)
	{
		//take individual PRSDs
		//printf("$dest: %d$\n", dest);
		cur_prsd = iter->getNext();
		if(!cur_prsd) break;	
		position = 0; 
		MPI_Pack(&msg_type, 1, MPI_UNSIGNED, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		code = 0xFFFF & cur_prsd->code;
		MPI_Pack(&code, 1, MPI_SHORT, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&cur_prsd->attr, 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&(cur_prsd->det_len), 1, MPI_SHORT, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);

		//details
		for(i=0;i<(cur_prsd->det_len);i++)
		{
			MPI_Pack(&(cur_prsd->det[i].type), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
			/* intra task info */
			MPI_Pack(&(cur_prsd->det[i].start_value), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
			MPI_Pack(&(cur_prsd->det[i].length), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
			MPI_Pack(&(cur_prsd->det[i].stride), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
			/* inter task info */
			MPI_Pack(&(cur_prsd->det[i].tid_start_value), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
			MPI_Pack(&(cur_prsd->det[i].tid_length), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
			MPI_Pack(&(cur_prsd->det[i].tid_stride), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
			/* inter node info */
			MPI_Pack(&(cur_prsd->det[i].node_id_start_value), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
			MPI_Pack(&(cur_prsd->det[i].node_id_length), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
			MPI_Pack(&(cur_prsd->det[i].node_id_stride), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);

			MPI_Pack(&(cur_prsd->det[i].loopsize), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
			MPI_Pack(&(cur_prsd->det[i].loopcount), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
			MPI_Pack(&(cur_prsd->det[i].eprsd_merge_count), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);		
			MPI_Pack(&(cur_prsd->det[i].tid_bit_pattern), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
			MPI_Pack(&(cur_prsd->det[i].nodeid_bit_pattern), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);		
			MPI_Pack(&(cur_prsd->det[i].tid_addr_stride), 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		}

		//common part
		MPI_Pack(&((cur_prsd->cmninfo)->max_sign_len), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&((cur_prsd->cmninfo)->sign_len), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		for(i=0;i<(cur_prsd->cmninfo)->sign_len;i++)
		{
			MPI_Pack(&((cur_prsd->cmninfo)->signature[i]), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);		
		}
		MPI_Pack(&((cur_prsd->cmninfo)->ref_id), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&((cur_prsd->cmninfo)->type), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&((cur_prsd->cmninfo)->thread_id), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(&((cur_prsd->cmninfo)->node_id), 1, MPI_UNSIGNED_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		MPI_Pack(((cur_prsd->cmninfo)->name), 10, MPI_CHAR, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);

		//user-defined part
		value = (cur_prsd->userinfo)->getValue();
		MPI_Pack(&value, 1, MPI_LONG, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
		//fprintf(STDOUT, "\nBefore send hi hi hi");
		//fflush(stdout);
		MPI_Send(sndBuffer, BUF_LEN, MPI_PACKED, dest, MSG_TAG, MPI_COMM_WORLD);
		count++;
		if(count%1000 == 0)
		{
			fprintf(STDOUT, "rank: %10d, %10d EPRSDs sent\r", myrank, count);	
		}

	}
	printf("\n***All PRSDs sent from %d to %d\n", myrank, dest);
	fflush(STDOUT);	
	return;
}

void sendPRSDsEndMsg_IN(int dest)
{
	char sndBuffer[BUF_LEN];
	int position = 0;
	int msg_type = CC_PRSDS_END_IN;
	MPI_Pack(&msg_type, 1, MPI_UNSIGNED, sndBuffer, BUF_LEN, &position, MPI_COMM_WORLD);
	MPI_Send(sndBuffer, BUF_LEN, MPI_PACKED, dest, MSG_TAG, MPI_COMM_WORLD);    
	return;
}

class mycompressor
{
	public:	
		EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE> *ecr;
		EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE>::ITERATOR *iter;
		COMMON* cmn;
		MYINFO* address;	
		_UINT64 sndSignArr[MAX_SIGN_LEN];

		mycompressor(int myrank)
		{
			ecr = new EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE>((_UINT64)(0x00000000FFFFFFFF&myrank));
			iter = new EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE>::ITERATOR(ecr);
			cmn = new COMMON(MAX_SIGN_LEN);
			address = new MYINFO();
		}

		~mycompressor()
		{
			delete address;
			delete cmn;
			delete iter;
			delete ecr;		
		}
};

mycompressor* cmprsr[MAX_THREADS];

int doIntraThreadCompression(int args)
{
	int num = 0, dest = 0, n_processes = 0;
	int myrank = args;
	char lt, gt;	
	_UINT64 val = 0;
	_UINT64 i = 0, sig_len = 0;  	
	FILE* fptr = NULL;
	_UINT64 tracecnt = MAX_TRACES;	

	printf("IntraThreadCompression starting\n");
	MPI_Comm_size(MPI_COMM_WORLD, &n_processes); 
	printf("no. of processes: %d\n", n_processes);

	while(tracecnt--)
		//while(1)
	{
		num = fscanf(fpipe,"\n*%ld*", &sign_size);
		//printf("sign_size = %ld", sign_size);getchar();
		if(num == 0 || num == -1){printf("\nEOF reached or ERROR#1....%d\n",num);goto TRACES_END;}
		sig_len = sign_size;
		if(sig_len > MAX_SIGN_LEN || sig_len < 0)
		{ 
			printf("\nSign length %d exceeded the limit %d, exiting", sig_len, MAX_SIGN_LEN);
			sig_len = MAX_SIGN_LEN;
		}
		num = fscanf(fpipe,"\n%c", &lt);
		if(num == 0 || num == -1){printf("\nEOF reached or ERROR#2....%d\n",num);goto TRACES_END;}
		for(i = 0; i < sig_len; i++)
		{
			fscanf(fpipe, " %20p", &val);
			if(num == 0 || num == -1){printf("\nEOF reached or ERROR#3....%d\n",num);goto TRACES_END;}
			/* ignore top most level of sign */
			//if(i)
			{	sndSignArr[i] = val;	}
			//fprintf(stdout, "% 20p", (void*)val);
		}
		fscanf(fpipe, "%c\n", &gt);
		if(num == 0 || num == -1){printf("\nEOF reached or ERROR#4....%d\n",num);goto TRACES_END;}

		num = fscanf(fpipe,"ip:%20p ebp:%20p sp:%20p sign: %20p tid:%4u %c %c addr:%20p %u\n",
				&ip,     &ebp,     &sp,     &sign,      &thread_id, &type, &valid, &addr, &sincelast);

		/*	
			fprintf(outfile,"ip:%20p ebp:%20p sp:%20p sign: %20p tid:%4u %c %c addr:%20p %u\n",
			ip,     ebp,     sp,     sign,      thread_id, type, valid, addr, sincelast);
			*/

TRACES_END:  
		if(num == 0 || num == -1)
		{
			printf("\nEOF reached or ERROR....%d\n",num);
			break;
		}

		//compress in this(main) process only, add to compressor object
		//TODO: enable later
		if(thread_id >= 0)
		{	
			if(cmprsr[thread_id] == NULL && thread_id < MAX_THREADS)
			{
				cmprsr[thread_id] = new mycompressor(myrank);
			}
			cmprsr[thread_id]->cmn->setData(sndSignArr, sign_size, sign, type, 
					0x00000000FFFFFFFF & thread_id, 
					0x00000000FFFFFFFF & myrank,		//TODO: node id change required 
					"LD_ST");
			cmprsr[thread_id]->address->setValue((void*)addr);
			cmprsr[thread_id]->ecr->add(cmprsr[thread_id]->cmn, cmprsr[thread_id]->address);
		}
		else
		{
			printf("Invalid thread id: %d in %d rank...exiting\n", thread_id, myrank);
			break;
		}


		if(tracecnt%10000 == 0)
		{
			fprintf(STDOUT, "%10d references remaining in process %10d\r", tracecnt, myrank);
		}

	}//while ends

	//MPI_Barrier(MPI_COMM_WORLD);

#if 0
	for(int i=0; i<MAX_THREADS; i++)
	{
		if(cmprsr[i])
		{
			printf("\n$$$$$$$$$finalize starting in %d$$$$$$$$$$$", myrank);
			cmprsr[i]->ecr->finalize();
			printf("\n$$$$$$$$$finalize completed in %d$$$$$$$$$$", myrank);
			//cmprsr[i]->ecr->finalize();
		}
	}
#endif	

	printf("\nIntrathread compression completed in %d", myrank);
	//MPI_Barrier(MPI_COMM_WORLD);

	return 0;
}

int doInterThreadCompression(int args)
{
	int myrank = args;
	EPRSD_DATA<MYINFO> tmpdata;
	EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE>::ITERATOR *iter = NULL;
	EPRSD_DATA<MYINFO> *cur_prsd = NULL;
	static int counter = 0;

	for(int stride = MAX_THREADS/2; stride > 0; stride /= 2)
	{
		printf("\nmyrank %ld - stride %d", myrank, stride);
		for(int dest = 0; dest < stride; dest++)
		{
			printf("\n$$$IT merge in %ld - merging EPRSDs from %d to %d thread$$$", myrank, dest+stride, dest);
			if(cmprsr[dest] == NULL || cmprsr[dest+stride] == NULL)
			{
				printf("\nNULL ptrs");
				break;
			}	
			/* merge dest+stride sigtree and EPRSDs to dest */
#ifdef SIGTREE_ENABLED
			printf("\nmerging sigtree - myrank %ld from %d to %d", myrank, dest+stride, dest);
			cmprsr[dest+stride]->ecr->numberSigTree(dest+stride, myrank); /* dest+stride is also tid */
			cmprsr[dest]->ecr->mergeSigTree(cmprsr[dest+stride]->ecr);
			printf("\nmerging sigtree completed - myrank %ld from %d to %d", myrank, dest+stride, dest);
			/* copy unique keys of leaf nodes to the dependent EPRSDs */
			cmprsr[dest+stride]->ecr->assignUniqueKeys();
#endif
			iter = new EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE>::ITERATOR(cmprsr[dest+stride]->ecr);

			cmprsr[dest]->ecr->newEPRSDSetBegins();
			if(!iter)
			{	printf("\n NULL iter");	return -2;		}

			iter->begin();
			printf("\nmerging EPRSDs - myrank %ld from %d to %d\n", myrank, dest+stride, dest);
			fflush(stdout);
			counter = 0;
			while(1)
			{
				cur_prsd = iter->getNext();
				if(!cur_prsd) break;
				tmpdata.copy( (CODE)(0xFFFF & cur_prsd->code), cur_prsd->attr, cur_prsd->det, cur_prsd->det_len, cur_prsd->cmninfo, cur_prsd->userinfo);
				cmprsr[dest]->ecr->interthread_add(&tmpdata);
				counter++;
				if(counter%100 == 0)
					printf("%d\r", counter);		
			}
			cmprsr[dest+stride]->ecr->~EPRSD_COMPRESSOR();
			cmprsr[dest]->ecr->interthread_startMerge();
			printf("\nmerging EPRSDs completed - myrank %ld from %d to %d", myrank, dest+stride, dest);
			delete iter;
			iter = NULL;
		}
	}

	if(cmprsr[1] == NULL)
	{
		printf("\nOnly one thread in the process %ld", myrank);
	}

	/* wait till all nodes finish inter-task compression */
	printf("\nrank %ld - intertask compression completed", myrank);
	MPI_Barrier(MPI_COMM_WORLD);
	return 0;
}

int doInterNodeCompression(int args)
{
	int myrank = 0, n_processes = -1;
	int internode_stride = 0;
	int terminate = 0;
	int first_entry = 0;
	int sender = 0;
	int position = 0;
	int msg_type = 0;
	MPI_Status status;
	CODE code;
	_INT16 det_len;
	_INT64 value = 0;
	_INT64 attr = ATTR_NONE;
	struct details det[3];
	EPRSD_DATA<MYINFO> tmpdata;
	COMMON* cmn = new COMMON(MAX_SIGN_LEN);
	MYINFO* address = new MYINFO();
	mycompressor* cur_cmprsr = NULL;

	myrank = args;
	MPI_Comm_size(MPI_COMM_WORLD, &n_processes);
	internode_stride = n_processes;
	cur_cmprsr = cmprsr[0];
	printf("\nInter-node merging starts in %d", myrank);

	terminate = 0;
	internode_stride = internode_stride/2;
	first_entry = 1;

	while(!terminate)
	{
		if(internode_stride == 0)
			break;

		if(myrank >= internode_stride) /* sender of EPRSDs */
		{
			sender = myrank - internode_stride;
			fprintf(STDOUT, "\nBefore receiving - if in %d............", myrank);
			/* wait for SIGTREE and EPRSD request messages from receiver */
			MPI_Recv(rcvBuffer, BUF_LEN, MPI_PACKED, 
					sender, //MPI_ANY_SOURCE, //sender, //src PID
					MSG_TAG, MPI_COMM_WORLD, &status);

			//unpack the buffer contents
			position = 0;
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &msg_type, 1, MPI_UNSIGNED, MPI_COMM_WORLD);

			switch(msg_type)
			{
#ifdef SIGTREE_ENABLED
				case CC_SIGTREE_REQ_IN:
					{
						printf("\nIN: CHILD %ld SIGTREE req msg rcvd from %d\n", myrank, sender);
						//send all signature tree nodes to sender of sigtree request msg
						SIGTREE_ITERATOR *titer = cur_cmprsr->ecr->createSigTreeIterator();
						if(!titer)
						{ printf("\nError: NULL sigtree iterator"); }
						sendSigtree_IN(sender, cur_cmprsr->ecr, titer, myrank, myrank);
						if(titer)
							cur_cmprsr->ecr->releaseSigTreeIterator(titer);
						//notify the end
						sendSigtreeEndMsg_IN(sender);
					}
					break;
#endif

				case CC_PRSDS_REQ_IN:
					{
						EPRSD_COMPRESSOR<MYINFO, MYMATCH, MYMERGE>::ITERATOR iter(cur_cmprsr->ecr);
						printf("\nIN: CHILD %ld PRSD req msg rcvd from %d\n", myrank, sender);
						//send all EPRSDs to sender of PRSD request msg
						sendAllPRSDs_IN(sender, cur_cmprsr->ecr, &iter);
						//notify the end
						sendPRSDsEndMsg_IN(sender);
						printf("\nIN: CHILD %ld all PRSDs sent to %d\n", myrank, sender);
						terminate = 1;	
					}
					break;
			}
		}
		else /* receiver of EPRSDs */
		{
			if(first_entry)
			{
				first_entry = 0;
				/* send SIGTREE and EPRSD request messages to sender */
				sender = myrank + internode_stride;
				printf("\n%ld IN - receiver, receiving from %d, sending PRSD/SIGTREE req msg", myrank, sender);
				/* before adding EPRSDs from a new node, inform compressor that new set of EPRSDs will be added */
				cur_cmprsr->ecr->newEPRSDSetBegins();
				/* inform sender that I am ready to receive EPRSDs - needed for synchronization in merging */
#ifdef SIGTREE_ENABLED
				sendSigtreeRequestMsg_IN(sender);
#else
				sendPRSDRequestMsg_IN(sender);
#endif
			}

			//fprintf(STDOUT, "\nBefore receiving - else in %d............", myrank);
			MPI_Recv(rcvBuffer, BUF_LEN, MPI_PACKED,
					sender, //MPI_ANY_SOURCE, //sender, //src PID
					MSG_TAG, MPI_COMM_WORLD, &status);

			//unpack the buffer contents
			position = 0;
			MPI_Unpack(rcvBuffer, BUF_LEN, &position, &msg_type, 1, MPI_UNSIGNED, MPI_COMM_WORLD);

			switch(msg_type)
			{
#ifdef SIGTREE_ENABLED
				case CC_SIGTREE_NODE_IN: /* signature tree's node received in response to CC_SIGTREE_REQ msg */
					{		
						SIGTREEITEM item;
						//printf("\nIN: CHILD %ld SIGTREE node rcvd from %d\n", myrank, sender);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(item.data.key[0]), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(item.data.key[1]), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(item.data.key[2]), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(item.data.info), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(item.parentData.key[0]), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(item.parentData.key[1]), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(item.parentData.key[2]), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(item.parentData.info), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						cur_cmprsr->ecr->mergeToSigTree(&item);
					}
					break;

				case CC_SIGTREE_END_IN:
					{
						printf("\nIN: CHILD %ld SIGTREE end msg rcvd from %d\n", myrank, sender);
						sendPRSDRequestMsg_IN(sender);
					}
					break;
#endif

				case CC_PRSDS_END_IN:
					internode_stride = internode_stride/2;
					first_entry = 1; /* make process to send SIGTREE and PRSD req msgs */
					printf("\nIN: CHILD %ld all prsds received\n", myrank);
					cur_cmprsr->ecr->startMerge(INTERNODE_MERGE);
					break;	

				case CC_PRSD_IN:
					{	
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &code, 1, MPI_SHORT, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &attr, 1, MPI_LONG, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &det_len, 1, MPI_SHORT, MPI_COMM_WORLD);
						//printf("\nIN: %ld received PRSD: attr = %d, det_len = %d", myrank, attr, det_len);
						//details
						for(_INT16 i=0;i<det_len && i<3;i++)
						{
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].type), 1, MPI_LONG, MPI_COMM_WORLD);
							/* intra task info */
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].start_value), 1, MPI_LONG, MPI_COMM_WORLD);
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].length), 1, MPI_LONG, MPI_COMM_WORLD);
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].stride), 1, MPI_LONG, MPI_COMM_WORLD);
							/* inter task info */
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].tid_start_value), 1, MPI_LONG, MPI_COMM_WORLD);
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].tid_length), 1, MPI_LONG, MPI_COMM_WORLD);
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].tid_stride), 1, MPI_LONG, MPI_COMM_WORLD);
							/* inter node info */
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].node_id_start_value), 1, 
									MPI_LONG, MPI_COMM_WORLD);
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].node_id_length), 1, MPI_LONG, MPI_COMM_WORLD);
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].node_id_stride), 1, MPI_LONG, MPI_COMM_WORLD);
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].loopsize), 1, MPI_LONG, MPI_COMM_WORLD);
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].loopcount), 1, MPI_LONG, MPI_COMM_WORLD);
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].eprsd_merge_count), 1, MPI_LONG, MPI_COMM_WORLD);
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].tid_bit_pattern), 1, 
									MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].nodeid_bit_pattern), 1, 
									MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(det[i].tid_addr_stride), 1, MPI_LONG, MPI_COMM_WORLD);
#if 0
							printf(" start_value = %p, length = %d, stride = %d, loopsize = %d, loopcount = %d, emc = %d", det[i].start_value, det[i].length, det[i].stride, det[i].loopsize, det[i].loopcount, det[i].eprsd_merge_count);
#endif
						}
						//common part
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->max_sign_len), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->sign_len), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);

						for(_UINT64 i=0;i<cmn->sign_len;i++)
						{
							MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->signature[i]), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						}
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->ref_id), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->type), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->thread_id), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->node_id), 1, MPI_UNSIGNED_LONG, MPI_COMM_WORLD);
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &(cmn->name), 10, MPI_CHAR, MPI_COMM_WORLD);

						//user-defined part
						MPI_Unpack(rcvBuffer, BUF_LEN, &position, &value, 1, MPI_LONG, MPI_COMM_WORLD);
						address->setValue((void*)value);

						//printf(" sign_len = %d", cmn->sign_len);
						//printf(" value = %p", value);		
						//printf(" ref_id = %p", cmn->ref_id);		
						//printf(" type = %d", cmn->type);		
						//printf(" tid = %d", cmn->thread_id);		
						//printf(" nid = %d", cmn->node_id);		
						//printf(" name = %s", cmn->name);
						//cmn->node_id = sender;//node id not available, hence not added during intra and inter-task compression		
						tmpdata.copy((CODE)(0xFFFF & code), attr, det, det_len, cmn, address);
						cur_cmprsr->ecr->internode_add(&tmpdata); 
					}	
					//receive from myrank+stride process
					break;

				default:
					{
						printf("\nError: Invalid message type, exiting, rank = %d", myrank);
						terminate = 1;
					}
					break;				
			}//switch() ends
		}//else ends
	}//while(1) ends

	MPI_Barrier(MPI_COMM_WORLD);
	printf("\nrank %ld - internode compression completed", myrank);
	return 0;
}

int doCleanup(int args)
{
	int myrank = args;

	if(myrank == 0)
	{	
		for(int i=0; i<MAX_THREADS; i++)
		{
			if(cmprsr[i] && i == 0)
			{
				cmprsr[i]->ecr->print(outfile);			
			}
		}
	}

	for(int i=0; i<MAX_THREADS; i++)
	{
		if(cmprsr[i])
		{
			delete cmprsr[i];
		}
	}
	MPI_Barrier(MPI_COMM_WORLD);
	return 0;
}

void startCompression(int myrank)
{
	int n_processes = -1;
	char outstr[16];	
	MPI_Comm_size(MPI_COMM_WORLD, &n_processes); 
	printf("no. of processes: %d\n", n_processes);
	printf("ROOT %d starting...\n", myrank);  
	//if((fpipe = fopen(PRSD_PIPE, "r"))== NULL)

	/* file or pipe to read trace from */
	sprintf(tracefilename, "trace%d.txt", myrank);
	printf("\n%d opening %s trace file/pipe", myrank, tracefilename);

	if((fpipe = fopen(tracefilename, "r"))== NULL)
	{
		printf("file/pipe %s not present....:!exiting\n", PRSD_PIPE);
		return;
	}

	/* only rank 0 and rank 1 create output files  */
	//if(myrank == 1 || myrank == 0)
	{	
		/* output file to store EPRSDs */
		sprintf(outstr, "prsd%d.txt", myrank);	
		if((outfile = fopen(outstr, "w")) == NULL)
		{
			printf("outfile open error in %d process....exiting\n", myrank);
			return;
		}
	}

	doIntraThreadCompression(myrank);
	gettimeofday(&t2, NULL);
	elapsed_time = (t2.tv_sec * 1000000 + t2.tv_usec) - (t1.tv_sec * 1000000 + t1.tv_usec);
	printf("\nIntra-thread elapsed time = %ld usec\n", elapsed_time);
	doInterThreadCompression(myrank);
	gettimeofday(&t2, NULL);
	elapsed_time = (t2.tv_sec * 1000000 + t2.tv_usec) - (t1.tv_sec * 1000000 + t1.tv_usec);
	printf("\nInter-thread elapsed time = %ld usec\n", elapsed_time);
	doInterNodeCompression(myrank);
	gettimeofday(&t2, NULL);
	elapsed_time = (t2.tv_sec * 1000000 + t2.tv_usec) - (t1.tv_sec * 1000000 + t1.tv_usec);
	printf("\nInter-node elapsed time = %ld usec\n", elapsed_time);
	doCleanup(myrank);

	fclose(outfile);
	fclose(fpipe);		
	return;
}


//starting point of the program
int main(int argc, char** argv)
{
	int myrank = 0, provided = 0;
	gettimeofday(&t1, NULL);
	MPI_Init( &argc, &argv );
	MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
	compstatus = STATUS_INTRATASK_COMP_IN_PROGRESS;
	startCompression(myrank);
	printf("\nrank %d terminating", myrank);	 
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Finalize(); 
	gettimeofday(&t2, NULL);
	elapsed_time = (t2.tv_sec * 1000000 + t2.tv_usec) - (t1.tv_sec * 1000000 + t1.tv_usec);
	printf("\nprsd_comp elapsed time = %ld usec\n", elapsed_time);
	return 0;
}
