/*
 *  This file contains the implementation of Signature Tree
 *  Created on Jan 01 2010
 *  Author: Sandeep Budanur Ramanna
 *  E-mail: sbudanu@ncsu.edu
 *  North Carolina State University
 */
#ifndef __SIGT_CPP
#define __SIGT_CPP

/* feature for faster tree building, tree is not searched for every child. Only the first child searches the whole tree to find the parent. Following 
   !!!conseutive!!! child nodes capitalize on the first search. If the sequence is broken, tree must be searched from the beginning */
#define PREV_PARENT_CMP

#define KEY_INVALID 0xFFFFFFFFFFFFFFFF

#include <stdio.h>
#include "types.h"
#include "myqueue.cpp"
#include "math.h"

typedef struct SIGNODE SIGNODE;

class SIGTREE_ITERATOR;

typedef struct SIGDATA
{
	_UINT64 key[3];
	void* info;
}SIGDATA;

struct SIGNODE
{
	char type;
	SIGDATA data;
	LIST<SIGNODE*> *child;
	SIGNODE* parent;
};

typedef struct SIGTREEITEM
{
	SIGDATA data;
	SIGDATA parentData;
}SIGTREEITEM;

/* SIGTREE iterator maintains a stack of following data structure. */
typedef struct SIGTREEITERDATA
{
	SIGNODE *curnode;
	LIST_ITERATOR<SIGNODE*> *childiter;

}SIGTREEITERDATA;

class SIGTREE
{
	private:
		SIGNODE root;
		/* pointer to store the current position in tree at which adding PCs/RAs is stopped in between while adding signature list item by item*/
		SIGNODE *traverse; 
		_UINT64 newnodeCount;/* newly added nodes during intra-thread compression*/
		_UINT64 duplicatenodeCount;/* duplicate nodes encountered during normal intra-thread compression */
		_UINT64 newnodeCountMerge;/* newly added node during merging/inter-task compression */
		_UINT64 duplicatenodeCountMerge;/* duplicate nodes encountered during merging/inter-task compression */
		_UINT64 numLeafNodesMerge;/* number of leafnodes/signatures found during merging */
		_UINT64 numLeafNodesNumbering; /* no. of leaf nodes found during numbering */
		_UINT64 createIterCount;
		_UINT64 releaseIterCount;	

	private:
		void releaseMem(SIGNODE* me)
		{
			SIGNODE* chld = NULL;
			LIST_ITERATOR<SIGNODE*> *iter = NULL;

			if(!me)
				return;

			iter = (me->child)->createIterator();
			/* free the memory of child nodes first aka Depth First style. Then free the current node's memory */
			while(iter->next())
			{
				//cout<<"*";
				chld = iter->currentItem();
				//fprintf(stdout, " %20p", (void*)val);
				releaseMem(chld);
			}
			me->parent = NULL;
			(me->child)->releaseIterator(iter);
			//(me->child)->~LIST();
			delete me->child;

			if(me != &root)/* root node is statically allocated */
				delete me;

			return;
		}

		SIGNODE* searchChildren(SIGNODE* rt , void* val)
		{
			LIST<SIGNODE*>* childList = NULL;
			LIST_ITERATOR<SIGNODE*>* iter = NULL;
			SIGNODE* cur = NULL;
			bool found = false;

			if(!rt){cout<<"\nNULL rt"; return NULL;}

			childList = rt->child;
			iter = childList->createIterator();
			/* traverse the children list */
			while(iter->next())
			{
				cur = iter->currentItem();
				if((cur->data).info == val)
				{
					found = true;
					break; 
				}
				//return cur;
			}
			childList->releaseIterator(iter);

			if(found)
				return cur;
			else 
				return NULL;
		}

		int insert(SIGNODE* rt, LIST_RITERATOR<void*> *iter, SIGNODE** leaf)
		{
			void* val = 0;
			bool present = false;
			SIGNODE* child = NULL;

			if(!iter) return -1;

			present = iter->next();
			*leaf = rt;
			if(!present)
				return 0;
			else /* add child nodes */
			{
				val = iter->currentItem();
				//fprintf(stdout, "\ninsert %20p", (void*)val);
				child = searchChildren(rt, val);
				if(child)/* already present, let the tree grow downwards, need not add at this level */
				{
					//fprintf(stdout, "\nchild found");
					duplicatenodeCount++;
					insert(child, iter, leaf);
				}
				else /* add new child */
				{
					//fprintf(stdout, "\nchild not found");
					newnodeCount++;
					SIGNODE* newnode = NULL;
					newnode = new SIGNODE();
					newnode->data.info = val;
					newnode->parent = rt;
					newnode->type = 0;
					newnode->child = new LIST<SIGNODE*>(); /* create a list object though it is empty at this stage */
					(rt->child)->addTail(newnode);
					insert(newnode, iter, leaf);
				}
			}
			return 0;
		}

		/* inserts node by node - state is preserved in 'traverse' pointer variable. This function must be preceded by initAdding() 
		 * and followed by finish Adding(). Otherwise, results will be ***unpredictable*** */
		int insert(SIGNODE* rt, void* pc_ra, SIGNODE** leaf)
		{
			//bool present = false;
			SIGNODE* child = NULL;

			if(!rt) return -1;

			//fprintf(stdout, "insert %20p\n", (void*)val);
			child = searchChildren(rt, pc_ra);
			if(child)/* already present, let the tree grow downwards, need not add at this level */
			{
				duplicatenodeCount++;
				traverse = child;
			}
			else /* add new child */
			{
				SIGNODE* newnode = NULL;
				newnode = new SIGNODE();
				newnode->data.info = pc_ra;
				newnode->parent = rt;
				newnode->type = 0;
				newnode->child = new LIST<SIGNODE*>(); /* create a list object though it is empty at this stage */
				(rt->child)->addTail(newnode);
				traverse = newnode;
				newnodeCount++;
			}
			*leaf = traverse;
			return 0;
		}

		/* inserts node by node - state is preserved in 'traverse' pointer variable. This function must be preceded by initAdding() 
		 * and followed by finish Adding(). Otherwise, results will be ***unpredictable*** 
		 * Adds 'src' node to the tree if it doesn't exist already.
		 * */
		int insert(SIGNODE* rt, SIGNODE* src, SIGNODE** leaf)
		{
			//bool present = false;
			SIGNODE* child = NULL;

			if(!rt) return -1;

			//fprintf(stdout, "insert %20p\n", (void*)val);
			child = searchChildren(rt, src->data.info);
			if(child)/* already present, let the tree grow downwards, need not add at this level */
			{
				if(child->data.info != src->data.info)/* shouldn't occur ideally */
				{
					fprintf(stdout, "\nError: node info mismatch while signature tree merging");
				}
				child->data = src->data; /* copy info(pc or ra) and key */
				duplicatenodeCountMerge++;
				traverse = child;
			}
			else /* add new child */
			{
				SIGNODE* newnode = NULL;
				newnode = new SIGNODE();
				newnode->data = src->data; /* copy info(pc or ra) and key */
				newnode->parent = rt;
				newnode->type = 0;
				newnode->child = new LIST<SIGNODE*>(); /* create a list object though it is empty at this stage */
				(rt->child)->addTail(newnode);
				traverse = newnode;
				newnodeCountMerge++;
			}
			*leaf = traverse;
			return 0;
		}

	public:
		friend class SIGTREE_ITERATOR;
		SIGTREE_ITERATOR* createIterator();
		void releaseIterator(SIGTREE_ITERATOR* iter);
		void printIteratorStatistics();
		bool getSignatureFromLeaf(SIGNODE* leaf, void **sign, _UINT64 *len);

		SIGTREE()
		{
			root.type = 0;
			root.data.info = NULL; //carries Program Counter or return address
			root.child = new LIST<SIGNODE*>();
			root.parent = NULL;
			duplicatenodeCount = 0;
			newnodeCount = 0;
			duplicatenodeCountMerge = 0;
			newnodeCountMerge = 0;
			numLeafNodesMerge = 0;
			numLeafNodesNumbering = 0;
			traverse = NULL;
			createIterCount = 0;
			releaseIterCount = 0;
		}

		~SIGTREE()
		{
			//cout<<"\nsigtree dtor\n";
			//printIteratorStatistics();
			releaseMem(&root);
		}

		void* add(LIST<void*> *signList)
		{
			//void* val = 0;
			SIGNODE* leaf = NULL;
			LIST_RITERATOR<void*> *iter = NULL;
			iter = signList->createRIterator();
			insert(&root, iter, &leaf);
			//cout<<"\nadd to tree";
#if 0
			while(iter->next())
			{
				val = iter->currentItem();
				fprintf(stdout, " %20p", (void*)val);
			}
#endif	
			signList->releaseRIterator(iter);

			return leaf;
		}

		void initAdding()
		{
			traverse = &root;
		}

		void finishAdding()
		{
			traverse = NULL;
		}

		/* calls to this function must be preceded by initAdding() and followed by finishAdding(). 
		   Otherwise, results will be ***unpredictable*** 
		   e.g:
		   initAdding();
		   while(!empty){add(pc_ra);}
		   finishAdding();

*/
		void* add(void* pc_ra)
		{
			SIGNODE* leaf = NULL;
			insert(traverse, pc_ra, &leaf);
			return leaf; 	
		}

	private:
		void* add(SIGNODE* src)
		{
			SIGNODE* leaf = NULL;
			insert(traverse, src, &leaf);
			return leaf;
		}

	public:
		void show(FILE* outfptr, SIGNODE* me, int level)
		{
			int i = 0;
			SIGNODE* chld = NULL;
			LIST_ITERATOR<SIGNODE*> *iter = NULL;

			if(!me)
				return;

			iter = (me->child)->createIterator();
			for(i=0;i<level;i++)/* show different levels */
			{fprintf(outfptr, "\t");}
			//fprintf(outfptr, " %-20p\n", (void*)me->data.info);
			fprintf(outfptr, " %-20p(%ld, %ld, %ld)\n", (void*)me->data.info, me->data.key[0], me->data.key[1], me->data.key[2]);
			while(iter->next())
			{
				//cout<<"*";
				chld = iter->currentItem();
				show(outfptr, chld, level+1);
				//fprintf(stdout, " %20p", (void*)chld->info);
			}
			(me->child)->releaseIterator(iter);
			return;		
		}	

		void print(FILE* outfptr)
		{
			fprintf(outfptr, "\n\nSignature Tree");
			fprintf(outfptr, "\n-----------------\n");
			show(outfptr, &root, 0);
			return;
		}

		void number(SIGNODE* me, _UINT64 *num, _UINT64 tid, _UINT64 node_id, _UINT64 level)
		{
			//int i = 0;
			SIGNODE* chld = NULL;
			LIST_ITERATOR<SIGNODE*> *iter = NULL;

			if(!me)
				return;

			iter = (me->child)->createIterator();

			if(me->child->getSize() == 0)/* leaf node */
			{
				numLeafNodesNumbering++;
			}

			me->data.key[0] = (*num)++;
			me->data.key[1] = tid;//thread id			
			me->data.key[2] = node_id;//node id

#if 0
			for(i=0;i<level;i++)/* show different levels */
			{fprintf(stdout, "\t");}
			fprintf(stdout, " %-20p(%d, %d, %d)\n", (void*)me->data.info, me->data.key[0], me->data.key[1], me->data.key[2]);
#endif

			while(iter->next())
			{
				//cout<<"*";
				chld = iter->currentItem();
				number(chld, num, tid, node_id, level+1);
				//fprintf(stdout, " %20p", (void*)chld->info);
			}
			(me->child)->releaseIterator(iter);
			return;		
		}	

		/* reset the key value of all nodes to KEY_INVALID = 0xFFFFFFFFFFFFFFFF */
		void resetNode(SIGNODE* me)
		{
			//int i = 0;
			SIGNODE* chld = NULL;
			LIST_ITERATOR<SIGNODE*> *iter = NULL;

			if(!me)
				return;

			iter = (me->child)->createIterator();
			me->data.key[0] = KEY_INVALID;//node#
			me->data.key[1] = KEY_INVALID;//thread id			
			me->data.key[2] = KEY_INVALID;//node id

			while(iter->next())
			{
				chld = iter->currentItem();
				resetNode(chld);
			}
			(me->child)->releaseIterator(iter);
			return;		
		}	

		/* adds tree nodes to a list */	
		void addToList(SIGNODE* me, LIST<SIGTREEITEM>* list)
		{
			//int i = 0;
			SIGNODE* chld = NULL;
			LIST_ITERATOR<SIGNODE*> *iter = NULL;
			SIGTREEITEM item;

			if(!me)
				return;

			/* bytewise copy*/
			item.data = me->data;

			if(me->parent)
				item.parentData = me->parent->data;
			else
			{
				item.parentData.key[0] = 0;
				item.parentData.key[1] = 0;
				item.parentData.key[2] = 0;
				item.parentData.info = NULL;			
			}
			list->addTail(item);

			iter = (me->child)->createIterator();

			while(iter->next())
			{
				//cout<<"*";
				chld = iter->currentItem();
				addToList(chld, list);
				//fprintf(stdout, " %20p", (void*)chld->info);
			}
			(me->child)->releaseIterator(iter);
			return;
		}

		/* linearizes the tree into a list */
		LIST<SIGTREEITEM>* createList()
		{
			LIST<SIGTREEITEM> *list = new LIST<SIGTREEITEM>();
			addToList(&root, list);
			return list;
		}

		/* releases the memory allocated for list by createList() */
		void releaseList(LIST<SIGTREEITEM> *list)
		{
			delete list;
		}

	private:
		void searchTree(SIGDATA *data, SIGNODE **found)
		{
			searchTreeByKey(&root, data, found);
			return;
		}

		/* search tree by key, *found contains the matching node's address */
		void searchTreeByKey(SIGNODE* me, SIGDATA *data, SIGNODE** found)
		{
			if(!me)
				return;

			//fprintf(stdout, "\n%-20p(%d, %d, %d)\n", (void*)me->data.info, me->data.key[0], me->data.key[1], me->data.key[2]);
			if(me->data.key[0] == data->key[0] &&
					me->data.key[1] == data->key[1] &&
					me->data.key[2] == data->key[2])
			{
				//fprintf(stdout, "\nsearchByKey found #1");
				*found = me;
			}
			else
			{
				LIST_ITERATOR<SIGNODE*>* iter = NULL;
				SIGNODE* chld = NULL;
				iter = (me->child)->createIterator();
				*found = NULL; /* if found this will be non-NULL */
				//fprintf(stdout, "\nsearchByKey - Before while");
				while(iter->next())
				{
					//fprintf(stdout, "\nsearchByKey - inside while");
					chld = iter->currentItem();
					searchTreeByKey(chld, data, found);
					if(*found)
					{
						//fprintf(stdout, "\nsearchByKey - found #2");
						break;
					}
				}
				(me->child)->releaseIterator(iter);
			}
			return;
		}

	public:

		/* merges the data contained in 'item' to the existing tree. If nodes were already present, they are updated otherwise
		   new nodes are created. Performing merge operation for all nodes on an empty tree creates an exact COPY of the tree being merged. */
		void merge(SIGTREEITEM *item)
		{
			SIGNODE* parent = NULL;
			SIGNODE* child = NULL;
			SIGNODE* newnode = NULL;

			//if dummy root node, just copy the contents
			if(item->data.key[0] == 0)
			{
				//printf("\nsigtree::merge(): dummy root node case");
				(this->root).data = item->data;
			}
			else
			{
				/* search the parent using unique key */
				searchTree(&(item->parentData), &parent);
				if(parent)/* parent found */
				{
					/* check whether item's info is already present in one of its children */
					child = searchChildren(parent, item->data.info);
					if(child) /* if present, update the child with item's key and info*/
					{
						//printf("\nsigtree::merge(): child already present");
						child->data = item->data;
					}
					else /* add as a new child */
					{
						//printf("\nsigtree::merge(): new child added");
						newnode = new SIGNODE();
						newnode->data = item->data;
						newnode->type = 0;
						newnode->child = new LIST<SIGNODE*>(); /* create a list object though it is empty at this stage */
						(parent->child)->addTail(newnode);
						newnode->parent = parent;
					}	
				}
				else
				{
					printf("\n!!!Error: SIGTREE: merge(): Parent not found, Adding to dummy root");
					newnode = new SIGNODE();
					newnode->data = item->data;
					newnode->type = 0;
					newnode->child = new LIST<SIGNODE*>(); /* create a list object though it is empty at this stage */
					(this->root).child->addTail(newnode);
					newnode->parent = &(this->root);
				}
			}

			return;		
		}

		/* construct a signature tree using the list representation of tree */
		static SIGTREE* buildTree(LIST<SIGTREEITEM> *list)
		{
			LIST_ITERATOR<SIGTREEITEM> *iter = NULL;
			SIGTREEITEM item;
			SIGTREE *newtree = new SIGTREE(); 
			SIGNODE *newnode = NULL, *parent = NULL, *prevParent = NULL;

			iter = (list)->createIterator();
			/* read items from the list one by one */
			while(iter->next())
			{
				parent = NULL;
				item = iter->currentItem();
				//first node is the dummy root node, just copy the contents
				if(item.data.key[0] == 0)
				{
					//printf("\nAdding dummy root node");
					(newtree->root).data = item.data;
				}
				else
				{
					newnode = new SIGNODE();
					newnode->data = item.data;
					newnode->parent = NULL;
					newnode->type = 0;
					newnode->child = new LIST<SIGNODE*>(); /* create a list object though it is empty at this stage */

					/* find parent by key */
#ifdef PREV_PARENT_CMP
					/* with last found parent first - avoids searching the whole tree for every child node of the same parent */
					if(prevParent)
					{				
						if(prevParent->data.key[0] == item.parentData.key[0] &&
								prevParent->data.key[1] == item.parentData.key[1] &&
								prevParent->data.key[2] == item.parentData.key[2])
						{
							parent = prevParent;
							//printf("\n***prev parent match***");
						}
					}
#endif

					if(!parent)/* if previous parent didn't match or no previous parent */
						newtree->searchTree(&(item.parentData), &parent);

					if(parent)
					{
#if 0
						//printf("\nAdding to parent");
						//fprintf(stdout, "\n%-20p(%d, %d, %d)\n", (void*)parent->data.info, parent->data.key[0], parent->data.key[1], parent->data.key[2]);
#endif
						(parent->child)->addTail(newnode);
						newnode->parent = parent;
					}
					else //all nodes should find a parent, else block to catch bugs if any
					{
						printf("\n!!!Error: SIGTREE: buildTree: Parent not found, Adding to dummy root");
						(newtree->root).child->addTail(newnode);
						newnode->parent = &(newtree->root);
					}
#ifdef PREV_PARENT_CMP
					prevParent = parent;
#endif

				}
			}
			(list)->releaseIterator(iter);
			return newtree;
		}

		void numberTree(_UINT64* num, _UINT64 tid, _UINT64 node_id)
		{
			numLeafNodesNumbering = 0;
			number(&root, num, tid, node_id, 0);
			//cout<<"\nNo. leaf nodes found during numbering: "<<numLeafNodesNumbering<<"\n";
			return;
		}

		void resetKeys()
		{
			resetNode(&root);
		}

		inline _UINT64 getChildNodeCount()
		{
			return newnodeCount;
		}

		inline _UINT64 getDuplicateCount()
		{
			return duplicatenodeCount;
		}

		inline _UINT64 getChildNodeMergeCount()
		{
			return newnodeCountMerge;
		}

		inline _UINT64 getDuplicateMergeCount()
		{
			return duplicatenodeCountMerge;
		}

		void mergeTrees(SIGNODE* src, LIST<SIGNODE*>* stack)
		{
			//int i = 0;
			SIGNODE* chld = NULL;
			LIST_ITERATOR<SIGNODE*> *iter = NULL;
			//SIGTREEITEM item;

			if(!src || !stack)
				return;
			if(src->data.key[0] > 0)/* do not push dummy root node */
				stack->addTail(src);

			iter = (src->child)->createIterator();

			while(iter->next())
			{
				chld = iter->currentItem();
				mergeTrees(chld, stack);
			}
			(src->child)->releaseIterator(iter);

			if(!chld) /* leaf node, no child found */
			{		
				LIST_ITERATOR<SIGNODE*> *stackIter = NULL;
				SIGNODE* item = NULL;
				stackIter = stack->createIterator();
				//fprintf(stdout, "\nAdding stack contents");
				initAdding();
				while(stackIter->next())
				{
					item = stackIter->currentItem();
					add(item);
					//fprintf(stdout, "\n%-20p(%d, %d, %d)", (void*)item->data.info, item->data.key[0], item->data.key[1], item->data.key[2]);
					fprintf(stdout, "%20p ",(void*)item->data.info);
				}
				fprintf(stdout, "\n");
				finishAdding();
				stack->releaseIterator(stackIter);
				numLeafNodesMerge++;
			}

			if(src->data.key[0] > 0)/* dummy root node not pushed */
				stack->removeTail();
			return;
		}

		/* traverse the 'src' tree in pre-order and build a stack of the nodes visited. When leaf node is reached, add the contents of the stack
		 * to destination tree. Pop the stack and push another leaf node and repeat the same process till whole signature list is merged. 
		 * While merging the keys are copied from src tree to dest tree. This will be useful while searching the signature leaf nodes for the newly arriving
		 * EPRSDs from the node whose signature tree is identical to 'src'.
		 *
		 * Before merging of EPRSDs, signature tree from src to dest node is sent and src tree is *rebuilt* and *merged* on the dest node. Then EPRSDs
		 * from src node is sent to dest node and merged in the dest node.
		 * */
		void merge(SIGTREE *src)
		{
			LIST<SIGNODE*> *stack = new LIST<SIGNODE*>();
			//fprintf(stdout, "\nmerging tree\n");
			mergeTrees(&(src->root), stack);
			//fprintf(stdout, "\nNo. of Leaf Nodes found during merge: %d\n", numLeafNodesMerge);
			delete stack;
		}

		/* get the leaf node pointer in the merged signature tree matchig the virtual signature or key. 'data' contains the PC and KEY to be searched. */
		void getLeafNode(SIGDATA *data, SIGNODE** found)
		{		
			searchTree(data, found);		
			return;
		}
};

#if 1
class SIGTREE_ITERATOR
{
	private:
		LIST<SIGTREEITERDATA> *stack;
		SIGTREE *tree;

	public:
		/*   */
		SIGTREE_ITERATOR(SIGTREE *t)
		{
			//cout<<"\nSIGTREE_ITERATOR ctor";
			tree = t;
			stack = new LIST<SIGTREEITERDATA>();	
		}

		~SIGTREE_ITERATOR()
		{
			SIGTREEITERDATA top;
			//cout<<"\nSIGTREE_ITERATOR dtor";
			//TODO: remove iterators 
			while(stack->getSize())
			{
				//cout<<"\nSIGTREE_ITERATOR dtor #####";
				stack->getTail(top);	
				(top.curnode)->child->releaseIterator(top.childiter);/* release the memory before popping */
				stack->removeTail();
			}
			delete stack;
		}

		bool next()
		{		
			SIGTREEITERDATA data;
			SIGTREEITERDATA top;

			data.curnode = NULL;
			data.childiter = NULL;

			//cout<<"\nSIGTREE_ITERATOR next()";

			if(stack->getSize() == 0)/* fresh start, stack empty, push the dummy root node */
			{
				//cout<<"\nSIGTREE_ITERATOR: pushing root node = ";
				data.curnode = &(tree->root);
				data.childiter = (data.curnode)->child->createIterator();
				stack->addTail(data);				
			}
			else
			{
				//cout<<"\nSIGTREE_ITERATOR: else#1";
				do
				{
					/* go to next child of the node at Stack's top */
					stack->getTail(top);		
					if(top.childiter->next())
					{
						//cout<<"\nSIGTREE_ITERATOR: else#2";
						data.curnode = top.childiter->currentItem();/* pointer to child node */
						data.childiter = (data.curnode)->child->createIterator();
						stack->addTail(data);
						break;
					}
					else /* no more children present, pop from the stack */
					{
						//cout<<"\nSIGTREE_ITERATOR: else#3";
						(top.curnode)->child->releaseIterator(top.childiter);/* release the memory before popping */
						stack->removeTail();

						if(top.curnode == &(tree->root))/* iteration complete, root node popped at last */
							return false;
					}					
				}while(1);/* if nodes have run out of children keep popping them till you find one with a child*/
			}

			//cout<<"\nSIGTREE_ITERATOR next() returns true";
			return true;
		}

		SIGTREEITEM* currentItem(SIGTREEITEM *item)
		{
			SIGTREEITERDATA top;
			//cout<<"\nSIGTREE_ITERATOR currentItem()";
			stack->getTail(top);
			if(top.curnode)
			{
				item->data = (top.curnode)->data;
				if((top.curnode)->parent)	
					item->parentData = (top.curnode)->parent->data;
			}
			return item;
		}
};

SIGTREE_ITERATOR* SIGTREE::createIterator()
{
	createIterCount++;
	return new SIGTREE_ITERATOR(this);
}

void SIGTREE::releaseIterator(SIGTREE_ITERATOR* iter)
{
	releaseIterCount++;
	delete iter;
}

void SIGTREE::printIteratorStatistics()
{
	printf("\nSIGTREE ITERATOR STATISTICS");
	printf("\n========================================");
	printf("\nCreated: %ld times", createIterCount);
	printf("\nReleased: %ld times", releaseIterCount);
	printf("\nMemory leak: %ld", createIterCount - releaseIterCount);
}

/* returns the signature list in 'sign'(array of void pointers) and its length in '*len' pointed by the 'leaf' node*/	
bool SIGTREE::getSignatureFromLeaf(SIGNODE* leaf, void **sign, _UINT64 *len)
{
	SIGNODE* cur = NULL;
	_UINT64 count = 0, i = 0;
	void* tmp = NULL;
	_UINT64 offset = 0;
	double doffset = 0.0;
	/* check for error */
	if(!leaf || !sign || *len == 0)
		return false;

	cur = leaf;

	while(cur)
	{
		if(count >= *len) /* not enough space */
			return false;

		sign[count] = (cur->data).info;
		count++;

		if(cur->parent == &root)/* need not visit dummy root node */
			break;

		cur = cur->parent;
	}

	doffset = ceil(1.0*count/2);
	offset = (_UINT64)doffset;
	/* invert the signature array */
	for(i=0; i<count/2; i++)
	{
		tmp = sign[i];
		sign[i] = sign[i+offset];
		sign[i+offset] = tmp;
	}

	*len = count;		
	return true;
}
#endif

#endif
